@@ -1,23 +1,21 @@
 import cgi
+import contextlib
 import datetime
 import email.message
 import json as jsonlib
 import typing
 import urllib.request
-import warnings
 from collections.abc import MutableMapping
 from http.cookiejar import Cookie, CookieJar
 from urllib.parse import parse_qsl, quote, unquote, urlencode
 
-import chardet
 import rfc3986
 import rfc3986.exceptions
 
-from .__version__ import __version__
-from ._content_streams import ByteStream, ContentStream, encode
+from ._content import PlainByteStream, encode_request, encode_response
 from ._decoders import (
     SUPPORTED_DECODERS,
-    Decoder,
+    ContentDecoder,
     IdentityDecoder,
     LineDecoder,
     MultiDecoder,
@@ -26,6 +24,7 @@ from ._decoders import (
 from ._exceptions import (
     HTTPCORE_EXC_MAP,
     CookieConflict,
+    DecodingError,
     HTTPStatusError,
     InvalidURL,
     NotRedirectResponse,
@@ -37,16 +36,19 @@ from ._exceptions import (
 )
 from ._status_codes import codes
 from ._types import (
+    ByteStream,
     CookieTypes,
     HeaderTypes,
     PrimitiveData,
     QueryParamTypes,
+    RawURL,
+    RequestContent,
     RequestData,
     RequestFiles,
+    ResponseContent,
     URLTypes,
 )
 from ._utils import (
-    ElapsedTimer,
     flatten_queryparams,
     guess_json_utf,
     is_known_encoding,
@@ -59,8 +61,58 @@ from ._utils import (
 
 
 class URL:
-    def __init__(self, url: URLTypes = "", params: QueryParamTypes = None) -> None:
-        if isinstance(url, str):
+    """
+    url = httpx.URL("HTTPS://jo%40email.com:a%20secret@example.com:1234/pa%20th?search=ab#anchorlink")
+
+    assert url.scheme == "https"
+    assert url.username == "jo@email.com"
+    assert url.password == "a secret"
+    assert url.userinfo == b"jo%40email.com:a%20secret"
+    assert url.host == "example.com"
+    assert url.port == 1234
+    assert url.netloc == "example.com:1234"
+    assert url.path == "/pa th"
+    assert url.query == b"?search=ab"
+    assert url.raw_path == b"/pa%20th?search=ab"
+    assert url.fragment == "anchorlink"
+
+    The components of a URL are broken down like this:
+
+    https://jo%40email.com:a%20secret@example.com:1234/pa%20th?search=ab#anchorlink
+    [scheme][  username  ] [password] [  host  ][port][ path ] [ query ] [fragment]
+            [       userinfo        ] [    netloc    ][    raw_path    ]
+
+    Note that:
+
+    * `url.scheme` is normalized to always be lowercased.
+
+    * `url.host` is normalized to always be lowercased, and is IDNA encoded. For instance:
+       url = httpx.URL("http://中国.icom.museum")
+       assert url.host == "xn--fiqs8s.icom.museum"
+
+    * `url.userinfo` is raw bytes, without URL escaping. Usually you'll want to work with
+      `url.username` and `url.password` instead, which handle the URL escaping.
+
+    * `url.raw_path` is raw bytes of both the path and query, without URL escaping.
+      This portion is used as the target when constructing HTTP requests. Usually you'll
+      want to work with `url.path` instead.
+
+    * `url.query` is raw bytes, without URL escaping. A URL query string portion can only
+      be properly URL escaped when decoding the parameter names and values themselves.
+    """
+
+    def __init__(
+        self, url: typing.Union["URL", str, RawURL] = "", params: QueryParamTypes = None
+    ) -> None:
+        if isinstance(url, (str, tuple)):
+            if isinstance(url, tuple):
+                raw_scheme, raw_host, port, raw_path = url
+                scheme = raw_scheme.decode("ascii")
+                host = raw_host.decode("ascii")
+                port_str = "" if port is None else f":{port}"
+                path = raw_path.decode("ascii")
+                url = f"{scheme}://{host}{port_str}{path}"
+
             try:
                 self._uri_reference = rfc3986.iri_reference(url).encode()
             except rfc3986.exceptions.InvalidAuthority as exc:
@@ -70,8 +122,12 @@ class URL:
                 # We don't want to normalize relative URLs, since doing so
                 # removes any leading `../` portion.
                 self._uri_reference = self._uri_reference.normalize()
-        else:
+        elif isinstance(url, URL):
             self._uri_reference = url._uri_reference
+        else:
+            raise TypeError(
+                f"Invalid type for url.  Expected str or httpx.URL, got {type(url)}"
+            )
 
         # Add any query parameters, merging with any in the URL if needed.
         if params:
@@ -85,73 +141,136 @@ class URL:
 
     @property
     def scheme(self) -> str:
+        """
+        The URL scheme, such as "http", "https".
+        Always normalised to lowercase.
+        """
         return self._uri_reference.scheme or ""
 
     @property
-    def authority(self) -> str:
-        port_str = self._uri_reference.port
-        default_port_str = {"https": "443", "http": "80"}.get(self.scheme, "")
-        if port_str is None or port_str == default_port_str:
-            return self._uri_reference.host or ""
-        return self._uri_reference.authority or ""
-
-    @property
-    def userinfo(self) -> str:
-        return self._uri_reference.userinfo or ""
+    def userinfo(self) -> bytes:
+        """
+        The URL userinfo as a raw bytestring.
+        For example: b"jo%40email.com:a%20secret".
+        """
+        userinfo = self._uri_reference.userinfo or ""
+        return userinfo.encode("ascii")
 
     @property
     def username(self) -> str:
+        """
+        The URL username as a string, with URL decoding applied.
+        For example: "jo@email.com"
+        """
         userinfo = self._uri_reference.userinfo or ""
         return unquote(userinfo.partition(":")[0])
 
     @property
     def password(self) -> str:
+        """
+        The URL password as a string, with URL decoding applied.
+        For example: "a secret"
+        """
         userinfo = self._uri_reference.userinfo or ""
         return unquote(userinfo.partition(":")[2])
 
     @property
     def host(self) -> str:
+        """
+        The URL host as a string.
+        Always normlized to lowercase, and IDNA encoded.
+
+        Examples:
+
+        url = httpx.URL("http://www.EXAMPLE.org")
+        assert url.host == "www.example.org"
+
+        url = httpx.URL("http://中国.icom.museum")
+        assert url.host == "xn--fiqs8s.icom.museum"
+        """
         return self._uri_reference.host or ""
 
     @property
     def port(self) -> typing.Optional[int]:
+        """
+        The URL port as an integer.
+        """
         port = self._uri_reference.port
         return int(port) if port else None
 
+    @property
+    def netloc(self) -> str:
+        """
+        Either `<host>` or `<host>:<port>` as a string.
+        Always normlized to lowercase, and IDNA encoded.
+        """
+        host = self._uri_reference.host or ""
+        port = self._uri_reference.port
+        return host if port is None else f"{host}:{port}"
+
     @property
     def path(self) -> str:
-        return self._uri_reference.path or "/"
+        """
+        The URL path as a string. Excluding the query string, and URL decoded.
+
+        For example:
+
+        url = httpx.URL("https://example.com/pa%20th")
+        assert url.path == "/pa th"
+        """
+        path = self._uri_reference.path or "/"
+        return unquote(path)
 
     @property
-    def query(self) -> str:
-        return self._uri_reference.query or ""
+    def query(self) -> bytes:
+        """
+        The URL query string, as raw bytes, excluding the leading b"?".
+        Note that URL decoding can only be applied on URL query strings
+        at the point of decoding the individual parameter names/values.
+        """
+        query = self._uri_reference.query or ""
+        return query.encode("ascii")
 
     @property
-    def full_path(self) -> str:
-        path = self.path
-        if self.query:
-            path += "?" + self.query
-        return path
+    def raw_path(self) -> bytes:
+        """
+        The complete URL path and query string as raw bytes.
+        Used as the target when constructing HTTP requests.
+
+        For example:
+
+        GET /users?search=some%20text HTTP/1.1
+        Host: www.example.org
+        Connection: close
+        """
+        path = self._uri_reference.path or "/"
+        if self._uri_reference.query is not None:
+            path += "?" + self._uri_reference.query
+        return path.encode("ascii")
 
     @property
     def fragment(self) -> str:
+        """
+        The URL fragments, as used in HTML anchors.
+        As a string, without the leading '#'.
+        """
         return self._uri_reference.fragment or ""
 
     @property
-    def raw(self) -> typing.Tuple[bytes, bytes, typing.Optional[int], bytes]:
+    def raw(self) -> RawURL:
+        """
+        The URL in the raw representation used by the low level
+        transport API. For example, see `httpcore`.
+
+        Provides the (scheme, host, port, target) for the outgoing request.
+        """
         return (
             self.scheme.encode("ascii"),
             self.host.encode("ascii"),
             self.port,
-            self.full_path.encode("ascii"),
+            self.raw_path,
         )
 
-    @property
-    def is_ssl(self) -> bool:
-        message = 'URL.is_ssl() is pending deprecation. Use url.scheme == "https"'
-        warnings.warn(message, DeprecationWarning)
-        return self.scheme == "https"
-
     @property
     def is_absolute_url(self) -> bool:
         """
@@ -166,36 +285,93 @@ class URL:
 
     @property
     def is_relative_url(self) -> bool:
+        """
+        Return `False` for absolute URLs such as 'http://example.com/path',
+        and `True` for relative URLs such as '/path'.
+        """
         return not self.is_absolute_url
 
     def copy_with(self, **kwargs: typing.Any) -> "URL":
-        if (
-            "username" in kwargs
-            or "password" in kwargs
-            or "host" in kwargs
-            or "port" in kwargs
-        ):
-            host = kwargs.pop("host", self.host)
-            port = kwargs.pop("port", self.port)
+        """
+        Copy this URL, returning a new URL with some components altered.
+        Accepts the same set of parameters as the components that are made
+        available via properties on the `URL` class.
+
+        For example:
+
+        url = httpx.URL("https://www.example.com").copy_with(username="jo@gmail.com", password="a secret")
+        assert url == "https://jo%40email.com:a%20secret@www.example.com"
+        """
+        allowed = {
+            "scheme": str,
+            "username": str,
+            "password": str,
+            "userinfo": bytes,
+            "host": str,
+            "port": int,
+            "netloc": str,
+            "path": str,
+            "query": bytes,
+            "raw_path": bytes,
+            "fragment": str,
+        }
+        for key, value in kwargs.items():
+            if key not in allowed:
+                message = f"{key!r} is an invalid keyword argument for copy_with()"
+                raise TypeError(message)
+            if value is not None and not isinstance(value, allowed[key]):
+                expected = allowed[key].__name__
+                seen = type(value).__name__
+                message = f"Argument {key!r} must be {expected} but got {seen}"
+                raise TypeError(message)
+
+        # Replace username, password, userinfo, host, port, netloc with "authority" for rfc3986
+        if "username" in kwargs or "password" in kwargs:
+            # Consolidate username and password into userinfo.
             username = quote(kwargs.pop("username", self.username) or "")
             password = quote(kwargs.pop("password", self.password) or "")
+            userinfo = f"{username}:{password}" if password else username
+            kwargs["userinfo"] = userinfo.encode("ascii")
 
-            authority = host
-            if port is not None:
-                authority += f":{port}"
-            if username:
-                userpass = username
-                if password:
-                    userpass += f":{password}"
-                authority = f"{userpass}@{authority}"
+        if "host" in kwargs or "port" in kwargs:
+            # Consolidate host and port into  netloc.
+            host = kwargs.pop("host", self.host) or ""
+            port = kwargs.pop("port", self.port)
+            kwargs["netloc"] = f"{host}:{port}" if port is not None else host
 
+        if "userinfo" in kwargs or "netloc" in kwargs:
+            # Consolidate userinfo and netloc into authority.
+            userinfo = (kwargs.pop("userinfo", self.userinfo) or b"").decode("ascii")
+            netloc = kwargs.pop("netloc", self.netloc) or ""
+            authority = f"{userinfo}@{netloc}" if userinfo else netloc
             kwargs["authority"] = authority
 
+        if "raw_path" in kwargs:
+            raw_path = kwargs.pop("raw_path") or b""
+            path, has_query, query = raw_path.decode("ascii").partition("?")
+            kwargs["path"] = path
+            kwargs["query"] = query if has_query else None
+
+        else:
+            # Ensure path=<url quoted str> for rfc3986
+            if kwargs.get("path") is not None:
+                kwargs["path"] = quote(kwargs["path"])
+
+            # Ensure query=<str> for rfc3986
+            if kwargs.get("query") is not None:
+                kwargs["query"] = kwargs["query"].decode("ascii")
+
         return URL(self._uri_reference.copy_with(**kwargs).unsplit())
 
     def join(self, url: URLTypes) -> "URL":
         """
         Return an absolute URL, using this URL as the base.
+
+        Eg.
+
+        url = httpx.URL("https://www.example.com/test")
+        url = url.join("/new/path")
+        assert url == "https://www.example.com/test/new/path"
         """
         if self.is_relative_url:
             return URL(url)
@@ -219,9 +395,10 @@ class URL:
         class_name = self.__class__.__name__
         url_str = str(self)
         if self._uri_reference.userinfo:
+            username = quote(self.username)
             url_str = (
                 rfc3986.urlparse(url_str)
-                .copy_with(userinfo=f"{self.username}:[secure]")
+                .copy_with(userinfo=f"{username}:[secure]")
                 .unsplit()
             )
         return f"{class_name}({url_str!r})"
@@ -239,7 +416,8 @@ class QueryParams(typing.Mapping[str, str]):
         value = args[0] if args else kwargs
 
         items: typing.Sequence[typing.Tuple[str, PrimitiveData]]
-        if value is None or isinstance(value, str):
+        if value is None or isinstance(value, (str, bytes)):
+            value = value.decode("ascii") if isinstance(value, bytes) else value
             items = parse_qsl(value)
         elif isinstance(value, QueryParams):
             items = value.multi_items()
@@ -340,13 +518,6 @@ class QueryParams(typing.Mapping[str, str]):
         query_string = str(self)
         return f"{class_name}({query_string!r})"
 
-    def getlist(self, key: typing.Any) -> typing.List[str]:
-        message = (
-            "QueryParams.getlist() is pending deprecation. Use QueryParams.get_list()"
-        )
-        warnings.warn(message, DeprecationWarning)
-        return self.get_list(key)
-
 
 class Headers(typing.MutableMapping[str, str]):
     """
@@ -572,48 +743,54 @@ class Headers(typing.MutableMapping[str, str]):
             return f"{class_name}({as_dict!r}{encoding_str})"
         return f"{class_name}({as_list!r}{encoding_str})"
 
-    def getlist(self, key: str, split_commas: bool = False) -> typing.List[str]:
-        message = "Headers.getlist() is pending deprecation. Use Headers.get_list()"
-        warnings.warn(message, DeprecationWarning)
-        return self.get_list(key, split_commas=split_commas)
-
-
-USER_AGENT = f"python-httpx/{__version__}"
-ACCEPT_ENCODING = ", ".join(
-    [key for key in SUPPORTED_DECODERS.keys() if key != "identity"]
-)
-
 
 class Request:
     def __init__(
         self,
-        method: str,
-        url: typing.Union[str, URL],
+        method: typing.Union[str, bytes],
+        url: typing.Union["URL", str, RawURL],
         *,
         params: QueryParamTypes = None,
         headers: HeaderTypes = None,
         cookies: CookieTypes = None,
+        content: RequestContent = None,
         data: RequestData = None,
         files: RequestFiles = None,
         json: typing.Any = None,
-        stream: ContentStream = None,
+        stream: ByteStream = None,
     ):
-        self.method = method.upper()
+        if isinstance(method, bytes):
+            self.method = method.decode("ascii").upper()
+        else:
+            self.method = method.upper()
         self.url = URL(url, params=params)
         self.headers = Headers(headers)
         if cookies:
             Cookies(cookies).set_cookie_header(self)
 
         if stream is not None:
+            # There's an important distinction between `Request(content=...)`,
+            # and `Request(stream=...)`.
+            #
+            # Using `content=...` implies automatically populated content headers,
+            # of either `Content-Length: ...` or `Transfer-Encoding: chunked`.
+            #
+            # Using `stream=...` will not automatically include any content headers.
+            #
+            # As an end-user you don't really need `stream=...`. It's only
+            # useful when:
+            #
+            # * Preserving the request stream when copying requests, eg for redirects.
+            # * Creating request instances on the *server-side* of the transport API.
             self.stream = stream
+            self._prepare({})
         else:
-            self.stream = encode(data, files, json)
+            headers, stream = encode_request(content, data, files, json)
+            self._prepare(headers)
+            self.stream = stream
 
-        self.timer = ElapsedTimer()
-        self.prepare()
-
-    def prepare(self) -> None:
-        for key, value in self.stream.get_headers().items():
+    def _prepare(self, default_headers: typing.Dict[str, str]) -> None:
+        for key, value in default_headers.items():
             # Ignore Transfer-Encoding if the Content-Length has been set explicitly.
             if key.lower() == "transfer-encoding" and "content-length" in self.headers:
                 continue
@@ -625,26 +802,16 @@ class Request:
         has_content_length = (
             "content-length" in self.headers or "transfer-encoding" in self.headers
         )
-        has_user_agent = "user-agent" in self.headers
-        has_accept = "accept" in self.headers
-        has_accept_encoding = "accept-encoding" in self.headers
-        has_connection = "connection" in self.headers
 
-        if not has_host:
-            url = self.url
-            if url.userinfo:
-                url = url.copy_with(username=None, password=None)
-            auto_headers.append((b"host", url.authority.encode("ascii")))
+        if not has_host and self.url.host:
+            default_port = {"http": 80, "https": 443}.get(self.url.scheme)
+            if self.url.port is None or self.url.port == default_port:
+                host_header = self.url.host.encode("ascii")
+            else:
+                host_header = self.url.netloc.encode("ascii")
+            auto_headers.append((b"host", host_header))
         if not has_content_length and self.method in ("POST", "PUT", "PATCH"):
             auto_headers.append((b"content-length", b"0"))
-        if not has_user_agent:
-            auto_headers.append((b"user-agent", USER_AGENT.encode("ascii")))
-        if not has_accept:
-            auto_headers.append((b"accept", b"*/*"))
-        if not has_accept_encoding:
-            auto_headers.append((b"accept-encoding", ACCEPT_ENCODING.encode()))
-        if not has_connection:
-            auto_headers.append((b"connection", b"keep-alive"))
 
         self.headers = Headers(auto_headers + self.headers.raw)
 
@@ -659,11 +826,12 @@ class Request:
         Read and return the request content.
         """
         if not hasattr(self, "_content"):
+            assert isinstance(self.stream, typing.Iterable)
             self._content = b"".join(self.stream)
             # If a streaming request has been read entirely into memory, then
             # we can replace the stream with a raw bytes implementation,
             # to ensure that any non-replayable streams can still be used.
-            self.stream = ByteStream(self._content)
+            self.stream = PlainByteStream(self._content)
         return self._content
 
     async def aread(self) -> bytes:
@@ -671,11 +839,12 @@ class Request:
         Read and return the request content.
         """
         if not hasattr(self, "_content"):
+            assert isinstance(self.stream, typing.AsyncIterable)
             self._content = b"".join([part async for part in self.stream])
             # If a streaming request has been read entirely into memory, then
             # we can replace the stream with a raw bytes implementation,
             # to ensure that any non-replayable streams can still be used.
-            self.stream = ByteStream(self._content)
+            self.stream = PlainByteStream(self._content)
         return self._content
 
     def __repr__(self) -> str:
@@ -689,29 +858,60 @@ class Response:
         self,
         status_code: int,
         *,
-        request: Request,
-        http_version: str = None,
         headers: HeaderTypes = None,
-        stream: ContentStream = None,
-        content: bytes = None,
+        content: ResponseContent = None,
+        text: str = None,
+        html: str = None,
+        json: typing.Any = None,
+        stream: ByteStream = None,
+        request: Request = None,
+        ext: dict = None,
         history: typing.List["Response"] = None,
+        on_close: typing.Callable = None,
     ):
         self.status_code = status_code
-        self.http_version = http_version
         self.headers = Headers(headers)
 
-        self.request = request
+        self._request: typing.Optional[Request] = request
+
         self.call_next: typing.Optional[typing.Callable] = None
 
+        self.ext = {} if ext is None else ext
         self.history = [] if history is None else list(history)
+        self._on_close = on_close
 
         self.is_closed = False
         self.is_stream_consumed = False
+
         if stream is not None:
-            self._raw_stream = stream
+            # There's an important distinction between `Response(content=...)`,
+            # and `Response(stream=...)`.
+            #
+            # Using `content=...` implies automatically populated content headers,
+            # of either `Content-Length: ...` or `Transfer-Encoding: chunked`.
+            #
+            # Using `stream=...` will not automatically include any content headers.
+            #
+            # As an end-user you don't really need `stream=...`. It's only
+            # useful when creating response instances having received a stream
+            # from the transport API.
+            self.stream = stream
         else:
-            self._raw_stream = ByteStream(body=content or b"")
-            self.read()
+            headers, stream = encode_response(content, text, html, json)
+            self._prepare(headers)
+            self.stream = stream
+            if content is None or isinstance(content, bytes):
+                # Load the response body, except for streaming content.
+                self.read()
+
+        self._num_bytes_downloaded = 0
+
+    def _prepare(self, default_headers: typing.Dict[str, str]) -> None:
+        for key, value in default_headers.items():
+            # Ignore Transfer-Encoding if the Content-Length has been set explicitly.
+            if key.lower() == "transfer-encoding" and "content-length" in self.headers:
+                continue
+            self.headers.setdefault(key, value)
 
     @property
     def elapsed(self) -> datetime.timedelta:
@@ -726,9 +926,32 @@ class Response:
             )
         return self._elapsed
 
+    @elapsed.setter
+    def elapsed(self, elapsed: datetime.timedelta) -> None:
+        self._elapsed = elapsed
+
+    @property
+    def request(self) -> Request:
+        """
+        Returns the request instance associated to the current response.
+        """
+        if self._request is None:
+            raise RuntimeError(
+                "The request instance has not been set on this response."
+            )
+        return self._request
+
+    @request.setter
+    def request(self, value: Request) -> None:
+        self._request = value
+
+    @property
+    def http_version(self) -> str:
+        return self.ext.get("http_version", "HTTP/1.1")
+
     @property
     def reason_phrase(self) -> str:
-        return codes.get_reason_phrase(self.status_code)
+        return self.ext.get("reason", codes.get_reason_phrase(self.status_code))
 
     @property
     def url(self) -> typing.Optional[URL]:
@@ -750,19 +973,22 @@ class Response:
             if not content:
                 self._text = ""
             else:
-                encoding = self.encoding
-                self._text = content.decode(encoding, errors="replace")
+                decoder = TextDecoder(encoding=self.encoding)
+                self._text = "".join([decoder.decode(self.content), decoder.flush()])
         return self._text
 
     @property
-    def encoding(self) -> str:
+    def encoding(self) -> typing.Optional[str]:
+        """
+        Return the encoding, which may have been set explicitly, or may have
+        been specified by the Content-Type header.
+        """
         if not hasattr(self, "_encoding"):
             encoding = self.charset_encoding
             if encoding is None or not is_known_encoding(encoding):
-                encoding = self.apparent_encoding
-                if encoding is None or not is_known_encoding(encoding):
-                    encoding = "utf-8"
-            self._encoding = encoding
+                self._encoding = None
+            else:
+                self._encoding = encoding
         return self._encoding
 
     @encoding.setter
@@ -778,40 +1004,25 @@ class Response:
         if content_type is None:
             return None
 
-        parsed = cgi.parse_header(content_type)
-        media_type, params = parsed[0], parsed[-1]
-        if "charset" in params:
-            return params["charset"].strip("'\"")
+        _, params = cgi.parse_header(content_type)
+        if "charset" not in params:
+            return None
 
-        # RFC 2616 specifies that 'iso-8859-1' should be used as the default
-        # for 'text/*' media types, if no charset is provided.
-        # See: https://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.7.1
-        if media_type.startswith("text/"):
-            return "iso-8859-1"
+        return params["charset"].strip("'\"")
 
-        return None
-
-    @property
-    def apparent_encoding(self) -> typing.Optional[str]:
-        """
-        Return the encoding, as it appears to autodetection.
-        """
-        return chardet.detect(self.content)["encoding"]
-
-    @property
-    def decoder(self) -> Decoder:
+    def _get_content_decoder(self) -> ContentDecoder:
         """
         Returns a decoder instance which can be used to decode the raw byte
         content, depending on the Content-Encoding used in the response.
         """
         if not hasattr(self, "_decoder"):
-            decoders: typing.List[Decoder] = []
+            decoders: typing.List[ContentDecoder] = []
             values = self.headers.get_list("content-encoding", split_commas=True)
             for value in values:
                 value = value.strip().lower()
                 try:
                     decoder_cls = SUPPORTED_DECODERS[value]
-                    decoders.append(decoder_cls(request=self.request))
+                    decoders.append(decoder_cls())
                 except KeyError:
                     continue
 
@@ -820,7 +1031,7 @@ class Response:
             elif len(decoders) > 1:
                 self._decoder = MultiDecoder(children=decoders)
             else:
-                self._decoder = IdentityDecoder(request=self.request)
+                self._decoder = IdentityDecoder()
 
         return self._decoder
 
@@ -841,12 +1052,19 @@ class Response:
             "For more information check: https://httpstatuses.com/{0.status_code}"
         )
 
+        request = self._request
+        if request is None:
+            raise RuntimeError(
+                "Cannot call `raise_for_status` as the request "
+                "instance has not been set on this response."
+            )
+
         if codes.is_client_error(self.status_code):
             message = message.format(self, error_type="Client Error")
-            raise HTTPStatusError(message, request=self.request, response=self)
+            raise HTTPStatusError(message, request=request, response=self)
         elif codes.is_server_error(self.status_code):
             message = message.format(self, error_type="Server Error")
-            raise HTTPStatusError(message, request=self.request, response=self)
+            raise HTTPStatusError(message, request=request, response=self)
 
     def json(self, **kwargs: typing.Any) -> typing.Any:
         if self.charset_encoding is None and self.content and len(self.content) > 3:
@@ -879,9 +1097,24 @@ class Response:
                 ldict[key] = link
         return ldict
 
+    @property
+    def num_bytes_downloaded(self) -> int:
+        return self._num_bytes_downloaded
+
     def __repr__(self) -> str:
         return f"<Response [{self.status_code} {self.reason_phrase}]>"
 
+    @contextlib.contextmanager
+    def _wrap_decoder_errors(self) -> typing.Iterator[None]:
+        # If the response has an associated request instance, we want decoding
+        # errors to be raised as proper `httpx.DecodingError` exceptions.
+        try:
+            yield
+        except ValueError as exc:
+            if self._request is None:
+                raise exc
+            raise DecodingError(message=str(exc), request=self.request) from exc
+
     def read(self) -> bytes:
         """
         Read and return the response content.
@@ -898,9 +1131,11 @@ class Response:
         if hasattr(self, "_content"):
             yield self._content
         else:
-            for chunk in self.iter_raw():
-                yield self.decoder.decode(chunk)
-            yield self.decoder.flush()
+            decoder = self._get_content_decoder()
+            with self._wrap_decoder_errors():
+                for chunk in self.iter_raw():
+                    yield decoder.decode(chunk)
+                yield decoder.flush()
 
     def iter_text(self) -> typing.Iterator[str]:
         """
@@ -908,18 +1143,20 @@ class Response:
         that handles both gzip, deflate, etc but also detects the content's
         string encoding.
         """
-        decoder = TextDecoder(request=self.request, encoding=self.charset_encoding)
-        for chunk in self.iter_bytes():
-            yield decoder.decode(chunk)
-        yield decoder.flush()
+        decoder = TextDecoder(encoding=self.encoding)
+        with self._wrap_decoder_errors():
+            for chunk in self.iter_bytes():
+                yield decoder.decode(chunk)
+            yield decoder.flush()
 
     def iter_lines(self) -> typing.Iterator[str]:
         decoder = LineDecoder()
-        for text in self.iter_text():
-            for line in decoder.decode(text):
+        with self._wrap_decoder_errors():
+            for text in self.iter_text():
+                for line in decoder.decode(text):
+                    yield line
+            for line in decoder.flush():
                 yield line
-        for line in decoder.flush():
-            yield line
 
     def iter_raw(self) -> typing.Iterator[bytes]:
         """
@@ -929,10 +1166,14 @@ class Response:
             raise StreamConsumed()
         if self.is_closed:
             raise ResponseClosed()
+        if not isinstance(self.stream, typing.Iterable):
+            raise RuntimeError("Attempted to call a sync iterator on an async stream.")
 
         self.is_stream_consumed = True
-        with map_exceptions(HTTPCORE_EXC_MAP, request=self.request):
-            for part in self._raw_stream:
+        self._num_bytes_downloaded = 0
+        with map_exceptions(HTTPCORE_EXC_MAP, request=self._request):
+            for part in self.stream:
+                self._num_bytes_downloaded += len(part)
                 yield part
         self.close()
 
@@ -956,8 +1197,8 @@ class Response:
         """
         if not self.is_closed:
             self.is_closed = True
-            self._elapsed = self.request.timer.elapsed
-            self._raw_stream.close()
+            if self._on_close is not None:
+                self._on_close(self)
 
     async def aread(self) -> bytes:
         """
@@ -975,9 +1216,11 @@ class Response:
         if hasattr(self, "_content"):
             yield self._content
         else:
-            async for chunk in self.aiter_raw():
-                yield self.decoder.decode(chunk)
-            yield self.decoder.flush()
+            decoder = self._get_content_decoder()
+            with self._wrap_decoder_errors():
+                async for chunk in self.aiter_raw():
+                    yield decoder.decode(chunk)
+                yield decoder.flush()
 
     async def aiter_text(self) -> typing.AsyncIterator[str]:
         """
@@ -985,18 +1228,20 @@ class Response:
         that handles both gzip, deflate, etc but also detects the content's
         string encoding.
         """
-        decoder = TextDecoder(request=self.request, encoding=self.charset_encoding)
-        async for chunk in self.aiter_bytes():
-            yield decoder.decode(chunk)
-        yield decoder.flush()
+        decoder = TextDecoder(encoding=self.encoding)
+        with self._wrap_decoder_errors():
+            async for chunk in self.aiter_bytes():
+                yield decoder.decode(chunk)
+            yield decoder.flush()
 
     async def aiter_lines(self) -> typing.AsyncIterator[str]:
         decoder = LineDecoder()
-        async for text in self.aiter_text():
-            for line in decoder.decode(text):
+        with self._wrap_decoder_errors():
+            async for text in self.aiter_text():
+                for line in decoder.decode(text):
+                    yield line
+            for line in decoder.flush():
                 yield line
-        for line in decoder.flush():
-            yield line
 
     async def aiter_raw(self) -> typing.AsyncIterator[bytes]:
         """
@@ -1006,10 +1251,14 @@ class Response:
             raise StreamConsumed()
         if self.is_closed:
             raise ResponseClosed()
+        if not isinstance(self.stream, typing.AsyncIterable):
+            raise RuntimeError("Attempted to call a async iterator on a sync stream.")
 
         self.is_stream_consumed = True
-        with map_exceptions(HTTPCORE_EXC_MAP, request=self.request):
-            async for part in self._raw_stream:
+        self._num_bytes_downloaded = 0
+        with map_exceptions(HTTPCORE_EXC_MAP, request=self._request):
+            async for part in self.stream:
+                self._num_bytes_downloaded += len(part)
                 yield part
         await self.aclose()
 
@@ -1032,8 +1281,8 @@ class Response:
         """
         if not self.is_closed:
             self.is_closed = True
-            self._elapsed = self.request.timer.elapsed
-            await self._raw_stream.aclose()
+            if self._on_close is not None:
+                await self._on_close(self)
 
 
 class Cookies(MutableMapping):
